/*
 * aserial.c - Routines for communicating (with an Adruino) using serial I/O
 * over USB.
*/

#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#ifdef __linux__
#include <linux/serial.h>
#endif
#include "aserial.h"

aserial_t *aserial_new(void)
{
	aserial_t *as;

	if ((as = malloc(sizeof(aserial_t))) == NULL)
		return NULL;
	as->fd = -1;
	as->msgno = 0;
	return as;
}

void aserial_destroy(aserial_t *as)
{
	if (as) free(as);
}

struct {
	int baud;
	speed_t mac;
} bds[] = {
	{ 50, B50 },
	{ 75, B75 },
	{ 110, B110 },
	{ 134, B134 },
	{ 150, B150 },
	{ 200, B200 },
	{ 300, B300 },
	{ 600, B600 },
	{ 1200, B1200 },
	{ 1800, B1800 },
	{ 2400, B2400 },
	{ 4800, B4800 },
#ifdef B7200
	{ 7200, B7200 },
#endif
	{ 9600, B9600 },
#ifdef B14400
	{ 14400, B14400 },
#endif
	{ 19200, B19200 },
#ifdef B28800
	{ 28800, B28800 },
#endif
	{ 38400, B38400 },
	{ 57600, B57600 },
#ifdef B76800
	{ 76800, B76800 },
#endif
	{ 115200, B115200 },
#ifdef B230400
	{ 230400, B230400 },
#endif
#ifdef B460800
	{ 460800, B460800 },
#endif
#ifdef B500000
	{ 500000, B500000 },
#endif
#ifdef B576000
	{ 576000, B576000 },
#endif
#ifdef B921600
	{ 921600, B921600 },
#endif
#ifdef B1000000
	{ 1000000, B1000000 },
#endif
#ifdef B1152000
	{ 1152000, B1152000 },
#endif
#ifdef B1500000
	{ 1500000, B1500000 },
#endif
#ifdef B2000000
	{ 2000000, B2000000 },
#endif
#ifdef B2500000
	{ 2500000, B2500000 },
#endif
#ifdef B3000000
	{ 3000000, B3000000 },
#endif
#ifdef B3500000
	{ 3500000, B3500000 },
#endif
#ifdef B4000000
	{ 4000000, B4000000 },
#endif
	{ 0, 0 }
};

static int aserial_set_baud(int fd, struct termios *term, int baud)
{
	int i = 0;

	while (bds[i].baud) {
		if (bds[i].baud == baud) {
			cfsetospeed(term, bds[i].mac);
			cfsetispeed(term, bds[i].mac);
			return 0;
		}
		i++;
	}
	return -1;
}

int aserial_open(aserial_t *as, char *name, int baud)
{
	int fd;
	struct termios term;
 	int bits;
#ifdef __linux__
	struct serial_struct	kernel_serial_settings;
#endif

	if ((fd = open(name, O_RDWR /*| O_NONBLOCK*/ | O_NOCTTY )) == -1) {
		fprintf(stderr, "open(%s) failed: %s\n", name, strerror(errno));
		return -1;
	}

	/* 
    * Reset the Arduino's line. This is key to getting the write to work.
    * Without it, the first few writes don't work.
    * Clear DTR, wait one second, flush input, then set DTR.
	* Without this, the first write fails.
 	*/
	if (ioctl(fd, TIOCMGET, &bits) < 0) {
		close(fd);
		perror("ioctl(TCIOMGET)");
		return -1;
	}
	bits &= ~(TIOCM_DTR | TIOCM_RTS);
	if (ioctl(fd, TIOCMSET, &bits) < 0) {
		close(fd);
		perror("ioctl(TIOCMSET)");
		return -1;
	}
#if 0
	sleep(1);
	tcflush(fd, TCIFLUSH);
	bits &= TIOCM_DTR;
	if (ioctl(fd, TIOCMSET, &bits) < 0) {
		close(fd);
		perror("ioctl(TIOCMSET)");
		return -1;
	}
#endif

	memset(&term, 0, sizeof(struct termios));
	term.c_iflag = IGNBRK | IGNPAR;
	term.c_cflag = CS8 | CREAD | HUPCL | CLOCAL;
	if (aserial_set_baud(fd, &term, baud) == -1) {
		close(fd);
		return -1;
	}
	if (tcsetattr(fd, TCSANOW, &term) < 0) {
		perror("tcsetattr()");
		return -1;
	}
#ifdef __linux__
	if (ioctl(fd, TIOCGSERIAL, &kernel_serial_settings) == 0) {
		kernel_serial_settings.flags |= ASYNC_LOW_LATENCY;
		ioctl(fd, TIOCSSERIAL, &kernel_serial_settings);
	}
#endif
	tcflush(fd, TCIFLUSH);
	as->fd = fd;
	return 0;
}

void aserial_close(aserial_t *as)
{
	if (as->fd != -1) {
		close(as->fd);
		as->fd = -1;
	}
}

/*
 * aserial_write - Writes to the serial port as "<" + data + ">".
 * data should be a NULL terminated string. nbytes is not used but
 * should be set for possible future use.
 *
 * Returns:
 * 0: some of the data wasn't written, try writing it all again
 * >0: should be the length of the NULL terminated data not including
 *     the NULL terminator
 * -1: error, errno is set
*/
ssize_t aserial_write(aserial_t *as, char *data, size_t nbytes)
{
	size_t nw;
	ssize_t n;
	char wbuf[100];

	// snprintf's length should include the null terminator
	snprintf(wbuf, 3+strlen(data)+2, "<%02d%s>", as->msgno, data);
	nw = strlen(wbuf);
#if UINTPTR_MAX >= 0xffffffffffffffff // 64-bit
printf("sending '%s', nw=%lu\n", wbuf, nw);
#else
printf("sending '%s', nw=%d\n", wbuf, nw);
#endif
	as->msgno++;
	if (as->msgno == 100) // keep to 2 digits
		as->msgno = 0;

	if ((n = write(as->fd, wbuf, nw)) == -1) {
		fprintf(stderr, "write() failed: %s\n", strerror(errno));
		return -1;
	}
	if (n < nw)
		return 0; // some not written, try again from scratch 
	return strlen(data); // wrote it all
}

ssize_t aserial_read(aserial_t *as, char *data, size_t nbytes)
{
	ssize_t n;

	while (1) {
		if ((n = read(as->fd, data, nbytes)) == -1) {
			if (errno != EAGAIN) {
				fprintf(stderr, "read() failed: (%d) %s\n", errno, strerror(errno));
				return -1;
			}
			printf("read() returned -1 and EAGAIN\n");
		} else {
			if (n > 0) {
				printf("\n");
				data[n] = '\0';
				printf("Got '%s'\n", data);
				return n;
			}
			// on Raspberry Pi, this printf seems to interfere with the
			// read, possibly by using up too much CPU
//			printf(".");
		}
	}
	return n;
}
