#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include "camcap.h"

static void print_caps(camcap_t *cc, __u32 caps);
static int print_cropcap_type(camcap_t *cc, __u32 type);
static void print_formats(camcap_t *cc);
static void print_video_inputs(camcap_t *cc);
static void print_controls(camcap_t *cc);

int main(int argc, char **argv)
{
	camcap_t *cc;

	if (argc != 2) {
		fprintf(stderr, "use: camcap_query <device_name>\n");
		exit(EXIT_FAILURE);
	}

	if ((cc = camcap_open(argv[1], NULL)) == NULL) {
		perror("camcap_open()");
		exit(EXIT_FAILURE);
	}

	printf("Capability info for %s\n", argv[1]);

	printf("driver: %s\n", cc->cap.driver);
	printf("card: %s\n", cc->cap.card);
	printf("bus_info: %s\n", cc->cap.bus_info);
	printf("version: %u.%u.%u\n",
		(cc->cap.version >> 16) & 0xff,
		(cc->cap.version >> 8) & 0xff,
		cc->cap.version & 0xff);

	printf("\ncapabilities (text from https://www.kernel.org/doc/html/v4.15/media/uapi/v4l/vidioc-querycap.html)\n");

	print_caps(cc, cc->cap.capabilities);
	if (cc->cap.capabilities & V4L2_CAP_DEVICE_CAPS) {
		printf("\ndevice_caps\n");
		print_caps(cc, cc->cap.device_caps);
	}

	print_formats(cc);
	print_video_inputs(cc);
	print_controls(cc);

	return EXIT_SUCCESS;
}

static void print_caps(camcap_t *cc, __u32 caps)
{
	if (caps & V4L2_CAP_VIDEO_CAPTURE) {
		printf("V4L2_CAP_VIDEO_CAPTURE 0x00000001 The device supports the single-planar API through the Video Capture interface.\n");
		print_cropcap_type(cc, V4L2_CAP_VIDEO_CAPTURE);
	}
	if (caps & V4L2_CAP_VIDEO_CAPTURE_MPLANE) {
		printf("V4L2_CAP_VIDEO_CAPTURE_MPLANE 0x00001000 The device supports the multi-planar API through the Video Capture interface.\n");
		print_cropcap_type(cc, V4L2_CAP_VIDEO_CAPTURE_MPLANE);
	}
	if (caps & V4L2_CAP_VIDEO_OUTPUT) {
		printf("V4L2_CAP_VIDEO_OUTPUT 0x00000002 The device supports the single-planar API through the Video Output interface.\n");
		print_cropcap_type(cc, V4L2_CAP_VIDEO_OUTPUT);
	}
	if (caps & V4L2_CAP_VIDEO_OUTPUT_MPLANE) {
		printf("V4L2_CAP_VIDEO_OUTPUT_MPLANE 0x00002000 The device supports the multi-planar API through the Video Output interface.\n");
		print_cropcap_type(cc, V4L2_CAP_VIDEO_OUTPUT_MPLANE);
	}
	if (caps & V4L2_CAP_VIDEO_M2M)
		printf("V4L2_CAP_VIDEO_M2M 0x00004000 The device supports the single-planar API through the Video Memory-To-Memory interface.\n");
	if (caps & V4L2_CAP_VIDEO_M2M_MPLANE)
		printf("V4L2_CAP_VIDEO_M2M_MPLANE 0x00008000 The device supports the multi-planar API through the Video Memory-To-Memory interface.\n");
	if (caps & V4L2_CAP_VIDEO_OVERLAY) {
		printf("V4L2_CAP_VIDEO_OVERLAY 0x00000004 The device supports the Video Overlay interface. A video overlay device typically stores captured images directly in the video memory of a graphics card, with hardware clipping and scaling.\n");
		print_cropcap_type(cc, V4L2_CAP_VIDEO_OVERLAY);
	}
	if (caps & V4L2_CAP_VBI_CAPTURE)
		printf("V4L2_CAP_VBI_CAPTURE 0x00000010 The device supports the Raw VBI Capture interface, providing Teletext and Closed Caption data.\n");
	if (caps & V4L2_CAP_VBI_OUTPUT)
		printf("V4L2_CAP_VBI_OUTPUT 0x00000020 The device supports the Raw VBI Output interface.\n");
	if (caps & V4L2_CAP_SLICED_VBI_CAPTURE)
		printf("V4L2_CAP_SLICED_VBI_CAPTURE 0x00000040 The device supports the Sliced VBI Capture interface.\n");
	if (caps & V4L2_CAP_SLICED_VBI_OUTPUT)
		printf("V4L2_CAP_SLICED_VBI_OUTPUT 0x00000080 The device supports the Sliced VBI Output interface.\n");
	if (caps & V4L2_CAP_RDS_CAPTURE)
		printf("V4L2_CAP_RDS_CAPTURE 0x00000100 The device supports the RDS capture interface.\n");
	if (caps & V4L2_CAP_VIDEO_OUTPUT_OVERLAY)
		printf("V4L2_CAP_VIDEO_OUTPUT_OVERLAY 0x00000200 The device supports the Video Output Overlay (OSD) interface. Unlike the Video Overlay interface, this is a secondary function of video output devices and overlays an image onto an outgoing video signal. When the driver sets this flag, it must clear the V4L2_CAP_VIDEO_OVERLAY flag and vice versa. [1]\n");
	if (caps & V4L2_CAP_HW_FREQ_SEEK)
		printf("V4L2_CAP_HW_FREQ_SEEK 0x00000400 The device supports the ioctl VIDIOC_S_HW_FREQ_SEEK ioctl for hardware frequency seeking.\n");
	if (caps & V4L2_CAP_RDS_OUTPUT)
		printf("V4L2_CAP_RDS_OUTPUT 0x00000800 The device supports the RDS output interface.\n");
	if (caps & V4L2_CAP_TUNER)
		printf("V4L2_CAP_TUNER 0x00010000 The device has some sort of tuner to receive RF-modulated video signals. For more information about tuner programming see Tuners and Modulators.\n");
	if (caps & V4L2_CAP_AUDIO)
		printf("V4L2_CAP_AUDIO 0x00020000 The device has audio inputs or outputs. It may or may not support audio recording or playback, in PCM or compressed formats. PCM audio support must be implemented as ALSA or OSS interface. For more information on audio inputs and outputs see Audio Inputs and Outputs.\n");
	if (caps & V4L2_CAP_RADIO)
		printf("V4L2_CAP_RADIO 0x00040000 This is a radio receiver.\n");
	if (caps & V4L2_CAP_MODULATOR)
		printf("V4L2_CAP_MODULATOR 0x00080000 The device has some sort of modulator to emit RF-modulated video/audio signals. For more information about modulator programming see Tuners and Modulators.\n");
	if (caps & V4L2_CAP_SDR_CAPTURE)
		printf("V4L2_CAP_SDR_CAPTURE 0x00100000 The device supports the SDR Capture interface.\n");
	if (caps & V4L2_CAP_EXT_PIX_FORMAT)
		printf("V4L2_CAP_EXT_PIX_FORMAT 0x00200000 The device supports the struct v4l2_pix_format extended fields.\n");
	if (caps & V4L2_CAP_SDR_OUTPUT)
		printf("V4L2_CAP_SDR_OUTPUT 0x00400000 The device supports the SDR Output interface.\n");
	if (caps & V4L2_CAP_META_CAPTURE)
		printf("V4L2_CAP_META_CAPTURE 0x00800000 The device supports the Metadata Interface capture interface.\n");
	if (caps & V4L2_CAP_READWRITE)
		printf("V4L2_CAP_READWRITE 0x01000000 The device supports the read() and/or write() I/O methods.\n");
	if (caps & V4L2_CAP_ASYNCIO)
		printf("V4L2_CAP_ASYNCIO 0x02000000 The device supports the asynchronous I/O methods.\n");
	if (caps & V4L2_CAP_STREAMING)
		printf("V4L2_CAP_STREAMING 0x04000000 The device supports the streaming I/O method.\n");
	if (caps & V4L2_CAP_TOUCH)
		printf("V4L2_CAP_TOUCH 0x10000000 This is a touch device.\n");
	if (caps & V4L2_CAP_DEVICE_CAPS)
		printf("V4L2_CAP_DEVICE_CAPS 0x80000000 The driver fills the device_caps field. This capability can only appear in the capabilities field and never in the device_caps field.\n");
}

static int print_cropcap_type(camcap_t *cc, __u32 type)
{
	struct v4l2_cropcap cr;
	struct v4l2_crop c;

	cr.type = type;
	if (ioctl(cc->fd, VIDIOC_CROPCAP, &cr) == -1) {
		if (errno == ENODATA) {
			printf("  Cropping (VIDIOC_CROPCAP) not supported\n");
			return 0;
		}
		perror("ioctl(VIDIOC_CROPCAP)");
		return -1;
	}
	printf("  Crop capabilities (VIDIOC_CROPCAP): (left, top, width, height in pixels)\n");
	printf("    bounds  %d, %d - %u, %u\n", cr.bounds.left, cr.bounds.top, cr.bounds.width, cr.bounds.height);
	printf("    defrect %d, %d - %u, %u\n", cr.defrect.left, cr.defrect.top, cr.defrect.width, cr.defrect.height);
	printf("    pixelaspect %u/%u (y/x)\n", cr.pixelaspect.numerator, cr.pixelaspect.denominator);

	switch (type) {
	case V4L2_CAP_VIDEO_CAPTURE:
		c.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		break;
	case V4L2_CAP_VIDEO_CAPTURE_MPLANE:
		c.type = V4L2_BUF_TYPE_VIDEO_CAPTURE_MPLANE;
		break;
	case V4L2_CAP_VIDEO_OUTPUT:
		c.type = V4L2_BUF_TYPE_VIDEO_OUTPUT;
		break;
	case V4L2_CAP_VIDEO_OUTPUT_MPLANE:
		c.type = V4L2_BUF_TYPE_VIDEO_OUTPUT_MPLANE;
		break;
	case V4L2_CAP_VIDEO_OVERLAY:
		c.type = V4L2_BUF_TYPE_VIDEO_OVERLAY;
		break;
	default:
		fprintf(stderr, "ioctl(VIDIOC_G_CROP): Invalid type\n");
		return -1;
	}	
	if (ioctl(cc->fd, VIDIOC_G_CROP, &c) == -1) {
		if (errno == EINVAL || errno == ENODATA) {
			printf("  Cropping (VIDIOC_G_CROP) not supported\n");
			//return 0;
		} else {
			perror("ioctl(VIDIOC_G_CROP)");
			//return -1;
		}
	} else {
		printf("  Current cropping rectangle (VIDIOC_G_CROP): (left, top, width, height in pixels)\n");
		printf("    %d, %d - %u, %u\n", c.c.left, c.c.top, c.c.width, c.c.height);
	}
	c.c.left = 10;
	c.c.top = 10;
	c.c.width = 100;
	c.c.height = 100;
	if (ioctl(cc->fd, VIDIOC_S_CROP, &c) == -1) {
		if (errno == EINVAL || errno == ENODATA) {
			printf("  Cropping (VIDIOC_S_CROP) not supported\n");
			return 0;
		}
		perror("ioctl(VIDIOC_S_CROP)");
		return -1;
	}
	printf("  Cropping (VIDIOC_S_CROP) supported\n");

	return 0;
}

static void print_formats(camcap_t *cc)
{
	// see https://www.kernel.org/doc/html/v4.15/media/uapi/v4l/vidioc-g-fmt.html#c.v4l2_format
	// https://www.kernel.org/doc/html/v4.15/media/uapi/v4l/pixfmt-v4l2.html#c.v4l2_pix_format
	struct v4l2_format fmt;
	struct v4l2_fmtdesc fd;

	memset(&fmt, 0, sizeof(fmt));
	fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	if (ioctl(cc->fd, VIDIOC_G_FMT, &fmt) == -1) {
		perror("VIDIOC_F_FMT");
		return;
	}
	printf("Formats:\n");
	printf("  width: %u, height: %u\n", fmt.fmt.pix.width, fmt.fmt.pix.height);
	printf("  sizeimage: %u bytes\n", fmt.fmt.pix.sizeimage);
	printf("  pixelformat: %s\n", fmt.fmt.pix.pixelformat == V4L2_PIX_FMT_YUYV ? "YUYV" : "other than YUYV");

	fd.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	fd.index = 0;
	while (1) {
		if (ioctl(cc->fd, VIDIOC_ENUM_FMT, &fd) == -1 ) {
			if (errno == EINVAL)
				break;
			perror("VIDIOC_ENUM_FMT");
			return;
		}
		printf("  %s\n", fd.description);
		printf("  flags: %x", fd.flags);
		if (fd.flags & V4L2_FMT_FLAG_COMPRESSED)
			printf(" V4L2_FMT_FLAG_COMPRESSED\n");
		if (fd.flags & V4L2_FMT_FLAG_EMULATED)
			printf(" V4L2_FMT_FLAG_EMULATED\n");
		printf("\n");
		printf("  See: https://www.kernel.org/doc/html/v4.15/media/uapi/v4l/pixfmt.html#pixfmt\n");
		fd.index++;
	}
}

static void print_video_inputs(camcap_t *cc)
{
	struct v4l2_input vi;

	printf("Video inputs:\n");
	vi.index = 0;
	while (1) {
		if (ioctl(cc->fd, VIDIOC_ENUMINPUT, &vi) == -1) {
			if (errno == EINVAL) {
				if (vi.index == 0)
					printf("  none\n");
				return; // done all video inputs
			}
			perror("ioctl(VIDIOC_ENUMINPUT)");
			return;
		}

		printf("  name: %s, type: ", vi.name);
		switch (vi.type) {
		case V4L2_INPUT_TYPE_TUNER:
			printf("V4L2_INPUT_TYPE_TUNER\n");
			break;
		case V4L2_INPUT_TYPE_CAMERA:
			printf("V4L2_INPUT_TYPE_CAMERA\n");
			break;
		case V4L2_INPUT_TYPE_TOUCH:
			printf("V4L2_INPUT_TYPE_TOUCH\n");
			break;
		default:
			printf("unknown\n");
		}

		printf("  standard: ");
		if (vi.std == 0)
			printf("usb\n");
		else
			printf("%llu (see https://www.kernel.org/doc/html/v4.15/media/uapi/v4l/vidioc-enumstd.html#v4l2-std-id)\n", vi.std);

		printf("  status: ");
		if (vi.status & V4L2_IN_ST_NO_POWER)
			printf("no power");
		else
			printf("on");
		if (vi.status & V4L2_IN_ST_NO_SIGNAL)
			printf(", no signal");
		else
			printf(", signal");

		printf(" capabilities: 0x%x\n", vi.capabilities);

		vi.index++;
	}
}

static void print_a_control(camcap_t *cc, __u32 id, char *name)
{
	struct v4l2_queryctrl queryctrl;

	queryctrl.id = id;
	if (ioctl(cc->fd, VIDIOC_QUERYCTRL, &queryctrl) == -1) {
		if (errno == EINVAL) {
			printf("no %s", name);
		} else {
			perror("VIDIOC_QUERYCTRL");
			return;
		}
	} else if (queryctrl.flags & V4L2_CTRL_FLAG_DISABLED) {
		printf("%s disabled", name);
	} else {
		printf("%s", name);
	}
}

/*
 * print_controls - Prints out some of the user controls. There
 * are many more. See the following for more and for how to
 * set them:
 * https://www.kernel.org/doc/html/v4.15/media/uapi/v4l/control.html
*/
static void print_controls(camcap_t *cc)
{
	struct v4l2_queryctrl queryctrl;

	printf("Controls: ");

	memset(&queryctrl, 0, sizeof(queryctrl));

	print_a_control(cc, V4L2_CID_BRIGHTNESS, "brightness");
	printf(", ");
	print_a_control(cc, V4L2_CID_AUTOBRIGHTNESS, "autobrightness");
	printf(", ");
	print_a_control(cc, V4L2_CID_CONTRAST, "contrast");
	printf(", ");
	print_a_control(cc, V4L2_CID_SATURATION, "saturation");
	printf(", ");
	print_a_control(cc, V4L2_CID_HUE, "hue");
	printf(", ");
	print_a_control(cc, V4L2_CID_GAMMA, "gamma");
	printf(", ");
	print_a_control(cc, V4L2_CID_GAIN, "gain");
	printf(", ");
	print_a_control(cc, V4L2_CID_AUTOGAIN, "autogain");
	printf(", ");
	print_a_control(cc, V4L2_CID_EXPOSURE, "exposure");
	printf("\n");
 	printf("  To change controls, see: https://www.kernel.org/doc/html/v4.15/media/uapi/v4l/control.html\n");
}
